
include("gauher.jl")
include("covariance_functions.jl")
include("gaussian_ops.jl")
include("utils.jl")

function laplace_moments(y, mu)
    lg = mu
    dlg = ones(size(mu,1))
    d2lg = zeros(size(mu,1))
    d3lg = zeros(size(mu,1))
    elg = exp(lg)
    lpi = lg.*y - elg - lgamma(y+1)
    dlpi = dlg * (y-elg)
    d2lpi = d2lg.*(y-elg) - dlg.*dlg.*elg
    return lpi, dlpi, d2lpi
end

function lik_ep_quad(y, mu, s2)
    #println("lik_ep_quad(%.3f,%.3f,%.3f)", y, mu, s2)
    N_quad = 20
    n = size(y,1)
    t,w = gauher(N_quad)
    lw = log(w')
    y = y*ones(n,1)
    mu = mu*ones(n,1)
    sig = sqrt(s2)*ones(n,1)

    # Laplace approximation
    lpi, dlpi, d2lpi = laplace_moments(y*ones(1,N_quad), sig*t'+mu*ones(1,N_quad))
    lZ,_ = logsumexp(lpi+lw)
    a = exp(lpi - lZ*ones(1,N_quad) + lw)
    dlZ = sum(a.*dlpi,2)
    d2lZ = sum(a.*(d2lpi+dlpi.*dlpi),2) - dlZ.*dlZ
    return lZ, dlZ, d2lZ
end

function bam_poisson_gp_inf(x_routine, x_events, x_events_flat, y_total, v, cov_routine, cov_events, fold; print_debug=true)
    DEBUG = false
    N = size(x_routine,1)
    D_routine = size(x_routine,2)
    D_events = size(x_events_flat,2)

    if print_debug
        println("\nrunning BAM-GP (Poisson)...")
    	println("iter 0: initialization")
    end

    # compute covariance matrix
    K_routine = covSEard(x_routine, x_routine, cov_routine["v"], cov_routine["l"])
    K_events = covSEard(x_events_flat, x_events_flat, cov_events["v"], cov_events["l"])

    # data structure for storing the approximate posterior on f
    BigNumber = 10.0^200
    q_fr = Dict("mu" => zeros(N,1), "v" => K_routine)
    q_fe = Dict("mu" => zeros(size(x_events_flat,1),1), "v" => K_events)

    # data structures for storing the approximate f_i marginals
    q_fir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    q_fie = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # data structures for storing the approximate y_i marginals
    q_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    q_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from f_i variables to g factor
    m_fir_gr = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_fir_gr_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_fie_ge = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_fie_ge_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from g factor to f_i variables
    m_gr_fir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_gr_fir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_ge_fie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_ge_fie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from f_i variables to h_i factors
    m_fir_hir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_fir_hir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_fie_hie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_fie_hie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from h_i factors to f_i variables
    m_hir_fir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hir_fir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hie_fie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_hie_fie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from y_i variables to h_i factors
    m_yir_hir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yir_hir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yie_hie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_yie_hie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from h_i factors to y_i variables
    m_hir_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hir_yir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hie_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_hie_yie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from y_i variables to k_i factor
    m_yir_ki = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yir_ki_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yie_ki = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_yie_ki_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from k_i factor to y_i variables
    m_ki_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_ki_yir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_ki_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_ki_yie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initialize messages and marginals approximations
    for i=1:N
        q_fir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1)*BigNumber)
        q_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1)*BigNumber)
        m_fir_gr[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_fir_gr_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_gr_fir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_gr_fir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_fir_hir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_fir_hir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_hir_fir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_hir_fir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_hir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_hir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_hir_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_hir_yir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_ki[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_ki_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_ki_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_ki_yir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        q_fie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        q_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_fie_ge_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_fie_ge_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_ge_fie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_ge_fie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_fie_hie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_fie_hie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_fie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_fie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_hie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_hie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_yie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_ki_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_ki_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_ki_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_ki_yie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        for e=1:size(x_events[i],1)
            push!(q_fie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1)*BigNumber))
            push!(q_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1)*BigNumber))
            push!(m_fie_ge_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_fie_ge_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_ge_fie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_ge_fie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_fie_hie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_fie_hie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_hie_fie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_hie_fie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_hie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_hie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_hie_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_hie_yie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_ki_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_ki_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_ki_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_ki_yie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
        end
        push!(q_fie, q_fie_inner)
        push!(q_yie, q_yie_inner)
        push!(m_fie_ge, m_fie_ge_inner)
        push!(m_fie_ge_old, m_fie_ge_old_inner)
        push!(m_ge_fie, m_ge_fie_inner)
        push!(m_ge_fie_old, m_ge_fie_old_inner)
        push!(m_fie_hie, m_fie_hie_inner)
        push!(m_fie_hie_old, m_fie_hie_old_inner)
        push!(m_hie_fie, m_hie_fie_inner)
        push!(m_hie_fie_old, m_hie_fie_old_inner)
        push!(m_yie_hie, m_yie_hie_inner)
        push!(m_yie_hie_old, m_yie_hie_old_inner)
        push!(m_hie_yie, m_hie_yie_inner)
        push!(m_hie_yie_old, m_hie_yie_old_inner)
        push!(m_yie_ki, m_yie_ki_inner)
        push!(m_yie_ki_old, m_yie_ki_old_inner)
        push!(m_ki_yie, m_ki_yie_inner)
        push!(m_ki_yie_old, m_ki_yie_old_inner)
    end

    # construct index mapping: [i][e] -> flatted_index e_ix
    index_mapping = Array{Int64}[]
    e_ix = 1
    for i=1:N
        inner_map = Array(Int64, size(x_events[i],1))
        for e=1:size(x_events[i],1)
            inner_map[e] = e_ix
            e_ix += 1
        end
        push!(index_mapping, inner_map)
    end

    # run EP
    max_iters = 40
    dampening_factor = 0.7
    yir_marginals = zeros(N,1)
    yie_marginals = Float64[]
    predictive_dist = Dict("mu" => zeros(N,1))
    lZi_hat_r = zeros(N,1)
    lZi_hat_e = zeros(size(x_events_flat,1),1)

    # tilde parameters
    tmu_r = zeros(N,1)
    tsigma_r = zeros(N,N)
    tmu_e = zeros(size(x_events_flat,1),1)
    tsigma_e = zeros(size(x_events_flat,1),size(x_events_flat,1))

    # cavity parameters
    nu_ni_r = zeros(N,1)
    tau_ni_r = zeros(N,1)
    nu_ni_e = zeros(size(x_events_flat,1),1)
    tau_ni_e = zeros(size(x_events_flat,1),1)

    N_quad = 20
    t,w = gauher(N_quad)
    lw = ones(1,1)*log(w')

    for iter=1:max_iters
        if print_debug
        	print("iter $iter: ")
        end
        
        yir_marginals = zeros(N,1)
        yie_marginals = Float64[]
        predictive_dist = Dict("mu" => zeros(N,1))

        for i=1:N # sweep through the data points
            if DEBUG println("processing data point $i, y($i)=$(y_total[i])"); end
            
            # step 1: compute the cavity distribution
            m_gr_fir[i] = gaussianDivision(Dict("mu" => q_fr["mu"][i]*ones(1,1), "v" => q_fr["v"][i,i]*ones(1,1)), m_fir_gr[i])
            #if m_gr_fir[i]["mu"][1,1] > 10
            #    m_gr_fir[i]["mu"][1,1] = m_gr_fir_old[i]["mu"][1,1]
            #    m_gr_fir[i]["v"][1,1] = m_gr_fir_old[i]["v"][1,1]
            #end
            m_gr_fir[i], m_gr_fir_old[i] = dampening(m_gr_fir[i], m_gr_fir_old[i], dampening_factor, iter)
            tau_ni_r[i] = 1.0 / m_gr_fir[i]["v"][1,1]
            nu_ni_r[i] = tau_ni_r[i] * m_gr_fir[i]["mu"][1,1]
            if DEBUG @printf("Step 1: cavity distribution R: m_gr_fir = N(%.3f, %.3f)\n", m_gr_fir[i]["mu"][1,1], m_gr_fir[i]["v"][1,1]); end
            for e=1:size(x_events[i],1)
                e_ix = index_mapping[i][e]
                m_ge_fie[i][e] = gaussianDivision(Dict("mu" => q_fe["mu"][e_ix]*ones(1,1), "v" => q_fe["v"][e_ix,e_ix]*ones(1,1)), m_fie_ge[i][e])
                #if m_ge_fie[i][e]["mu"][1,1] > 10
                #    m_ge_fie[i][e]["mu"][1,1] = m_ge_fie_old[i][e]["mu"][1,1]
                #    m_ge_fie[i][e]["v"][1,1] = m_ge_fie_old[i][e]["v"][1,1]
                #end
                m_ge_fie[i][e], m_ge_fie_old[i][e] = dampening(m_ge_fie[i][e], m_ge_fie_old[i][e], dampening_factor, iter)
                tau_ni_e[e_ix] = 1.0 / m_ge_fie[i][e]["v"][1,1]
                nu_ni_e[e_ix] = tau_ni_e[e_ix] * m_ge_fie[i][e]["mu"][1,1]
                if DEBUG @printf("Step 1: cavity distribution E%d: m_ge_fie = N(%.3f, %.3f)\n", e, m_ge_fie[i][e]["mu"][1,1], m_ge_fie[i][e]["v"][1,1]); end
            end
            
            
            # step 2: compute approximate posterior on f_i (site parameters)
            q_fir[i] = gaussianProduct(m_gr_fir[i], m_hir_fir[i])
            if DEBUG @printf("Step 2: marginal R: q_fir = N(%.3f, %.3f)\n", q_fir[i]["mu"][1,1], q_fir[i]["v"][1,1]); end
            for e=1:size(x_events[i],1)
                q_fie[i][e] = gaussianProduct(m_ge_fie[i][e], m_hie_fie[i][e])
                if DEBUG @printf("Step 2: marginal E%d: q_fie = N(%.3f, %.3f)\n", e, q_fie[i][e]["mu"][1,1], q_fie[i][e]["v"][1,1]); end
            end        
            
            
            # step 3: message from f_i variables to h_i factors
            m_fir_hir[i] = gaussianDivision(q_fir[i], m_hir_fir[i])
            m_fir_hir[i], m_fir_hir_old[i] = dampening(m_fir_hir[i], m_fir_hir_old[i], dampening_factor, iter)
            for e=1:size(x_events[i],1)
                m_fie_hie[i][e] = gaussianDivision(q_fie[i][e], m_hie_fie[i][e])
                m_fie_hie[i][e], m_fie_hie_old[i][e] = dampening(m_fie_hie[i][e], m_fie_hie_old[i][e], dampening_factor, iter)
            end
            
            
            # step 4: message from h_i factors to y_i variables
            #m_hir_yir[i] = gaussianMarginal(m_fir_hir[i], 1, 0, beta_routine)
            #m_hir_yir[i], m_hir_yir_old[i] = dampening(m_hir_yir[i], m_hir_yir_old[i], dampening_factor, iter)
            #for e=1:size(x_events[i],1)
            #    m_hie_yie[i][e] = gaussianMarginal(m_fie_hie[i][e], 1, 0, beta_events)
            #    m_hie_yie[i][e], m_hie_yie_old[i][e] = dampening(m_hie_yie[i][e], m_hie_yie_old[i][e], dampening_factor, iter)
            #end
            

            # step 4
            # first moment using Gaussian-Hermite quad
            mu = m_fir_hir[i]["mu"].*ones(1,1)
            sig = sqrt(m_fir_hir[i]["v"]).*ones(1,1)
            lg = sig*t'+mu*ones(1,N_quad)
            lZ,_ = logsumexp(lg+lw)
            ymu = exp(lZ)
            # second y moment from Poisson distribution
            elg = exp(lg)
            yv = elg
            ys2 = (yv+(elg-ymu*ones(1,N_quad)).^2)*w
            m_hir_yir[i] = Dict("mu" => ones(1,1) * ymu[1,1], "v" => ones(1,1) * ys2[1,1])
            m_hir_yir[i], m_hir_yir_old[i] = dampening(m_hir_yir[i], m_hir_yir_old[i], dampening_factor, iter)
            if DEBUG @printf("Step 4: predictive R: m_hir_yir = N(%.3f, %.3f)\n", m_hir_yir[i]["mu"][1,1], m_hir_yir[i]["v"][1,1]); end
            for e=1:size(x_events[i],1)
                # first moment using Gaussian-Hermite quad
                mu = m_fie_hie[i][e]["mu"].*ones(1,1)
                sig = sqrt(m_fie_hie[i][e]["v"]).*ones(1,1)
                lg = sig*t'+mu*ones(1,N_quad)
                lZ,_ = logsumexp(lg+lw)
                ymu = exp(lZ)
                # second y moment from Poisson distribution
                elg = exp(lg)
                yv = elg
                ys2 = (yv+(elg-ymu*ones(1,N_quad)).^2)*w
                m_hie_yie[i][e] = Dict("mu" => ones(1,1) * ymu[1,1], "v" => ones(1,1) * ys2[1,1])
                m_hie_yie[i][e], m_hie_yie_old[i][e] = dampening(m_hie_yie[i][e], m_hie_yie_old[i][e], dampening_factor, iter)
            if DEBUG @printf("Step 4: predictive E%d: m_hie_yie = N(%.3f, %.3f)\n", e, m_hie_yie[i][e]["mu"][1,1], m_hie_yie[i][e]["v"][1,1]); end
            end

            
            # step 5: compute posterior on y_i^r and y_i^e variables
            q_yir[i] = gaussianProduct(m_hir_yir[i], m_ki_yir[i])
            #q_yir[i], Z_hat = gaussianTruncation(q_yir[i], 0)
            if DEBUG @printf("Step 5: marginal R: q_yir = N(%.3f, %.3f)\n", q_yir[i]["mu"][1,1], q_yir[i]["v"][1,1]); end
            for e=1:size(x_events[i],1)
                q_yie[i][e] = gaussianProduct(m_hie_yie[i][e], m_ki_yie[i][e])
                #q_yie[i][e], Z_hat = gaussianTruncation(q_yie[i][e], 0)
                if DEBUG @printf("Step 5: marginal E%d: q_yie = N(%.3f, %.3f)\n", e, q_yie[i][e]["mu"][1,1], q_yie[i][e]["v"][1,1]); end
            end


            # step 6: message from y_i variables to k_i factor
            m_yir_ki[i] = gaussianDivision(q_yir[i], m_ki_yir[i])
            m_yir_ki[i], m_yir_ki_old[i] = dampening(m_yir_ki[i], m_yir_ki_old[i], dampening_factor, iter)
            for e=1:size(x_events[i],1)
                m_yie_ki[i][e] = gaussianDivision(q_yie[i][e], m_ki_yie[i][e])
                m_yie_ki[i][e], m_yie_ki_old[i][e] = dampening(m_yie_ki[i][e], m_yie_ki_old[i][e], dampening_factor, iter)
            end
            
            
            # step 7: message from k_i factor to y_i variables
            m_ki_yir[i]["mu"] = y_total[i]*ones(1,1)
            m_ki_yir[i]["v"] = v*ones(1,1)
            for e=1:size(x_events[i],1)
                m_ki_yir[i]["mu"] -= m_yie_ki[i][e]["mu"]
                m_ki_yir[i]["v"] += m_yie_ki[i][e]["v"]
            end
            m_ki_yir[i], m_ki_yir_old[i] = dampening(m_ki_yir[i], m_ki_yir_old[i], dampening_factor, iter)
            if DEBUG @printf("Step 7: residual R: m_ki_yir = N(%.3f, %.3f)\n", m_ki_yir[i]["mu"][1,1], m_ki_yir[i]["v"][1,1]); end
            for e=1:size(x_events[i],1)
                m_ki_yie[i][e]["mu"] = y_total[i] - m_yir_ki[i]["mu"]
                m_ki_yie[i][e]["v"] = v + m_yir_ki[i]["v"]
                for e2=1:size(x_events[i],1)
                    if e2 != e
                        m_ki_yie[i][e]["mu"] -= m_yie_ki[i][e2]["mu"]
                        m_ki_yie[i][e]["v"] += m_yie_ki[i][e2]["v"]
                    end
                end
                m_ki_yie[i][e], m_ki_yie_old[i][e] = dampening(m_ki_yie[i][e], m_ki_yie_old[i][e], dampening_factor, iter)
                if DEBUG @printf("Step 7: residual E%d: m_ki_yie = N(%.3f, %.3f)\n", e, m_ki_yie[i][e]["mu"][1,1], m_ki_yie[i][e]["v"][1,1]); end
            end
            
            
            # step 8: compute posterior on y_i variables
            q_yir[i] = gaussianProduct(m_hir_yir[i], m_ki_yir[i])
            #q_yir[i], Z_hat = gaussianTruncation(q_yir[i], 0)
            #print("Z_hat:", Z_hat, log(Z_hat))
            yir_marginals[i] = q_yir[i]["mu"][1,1]
            predictive_dist["mu"][i] = q_yir[i]["mu"][1,1]
            if DEBUG @printf("Step 8: marginal R: q_yir = N(%.3f, %.3f)\n", q_yir[i]["mu"][1,1], q_yir[i]["v"][1,1]); end
            for e=1:size(x_events[i],1)
                q_yie[i][e] = gaussianProduct(m_hie_yie[i][e], m_ki_yie[i][e])
                #q_yie[i][e], Z_hat = gaussianTruncation(q_yie[i][e], 0)
                #print("Z_hat:", Z_hat, log(Z_hat))
                push!(yie_marginals, q_yie[i][e]["mu"][1,1])
                predictive_dist["mu"][i] += q_yie[i][e]["mu"][1,1]
                if DEBUG @printf("Step 8: marginal E%d: q_yie = N(%.3f, %.3f)\n", e, q_yie[i][e]["mu"][1,1], q_yie[i][e]["v"][1,1]); end
            end
            
            
            # step 9: message from y_i variables to h_i factors
            m_yir_hir[i] = gaussianDivision(q_yir[i], m_hir_yir[i])
            m_yir_hir[i], m_yir_hir_old[i] = dampening(m_yir_hir[i], m_yir_hir_old[i], dampening_factor, iter)
            for e=1:size(x_events[i],1)
                m_yie_hie[i][e] = gaussianDivision(q_yie[i][e], m_hie_yie[i][e])
                m_yie_hie[i][e], m_yie_hie_old[i][e] = dampening(m_yie_hie[i][e], m_yie_hie_old[i][e], dampening_factor, iter)
            end
            
            
            # step 10: message from h_i factors to f_i variables
            #m_hir_fir[i] = gaussianMarginal(m_yir_hir[i], 1, 0, beta_routine)
            #m_hir_fir[i], m_hir_fir_old[i] = dampening(m_hir_fir[i], m_hir_fir_old[i], dampening_factor, iter)
            #for e=1:size(x_events[i],1)
            #    m_hie_fie[i][e] = gaussianMarginal(m_yie_hie[i][e], 1, 0, beta_events)
            #    m_hie_fie[i][e], m_hie_fie_old[i][e] = dampening(m_hie_fie[i][e], m_hie_fie_old[i][e], dampening_factor, iter)
            #end


            # steps 10, 11 & 12
            if DEBUG @printf("EP quad R: y=%.3f     m_yir_hir = N(%.3f, %.3f)\n", m_yir_hir[i]["mu"][1,1], m_gr_fir[i]["mu"][1,1], m_gr_fir[i]["v"][1,1]); end
            lZ, dlZ, d2lZ = lik_ep_quad(max(m_yir_hir[i]["mu"],0), m_gr_fir[i]["mu"], m_gr_fir[i]["v"])
            tau_ni = 1.0 / m_gr_fir[i]["v"][1,1]
            nu_ni = m_gr_fir[i]["mu"][1,1] / m_gr_fir[i]["v"][1,1]
            ttau = -d2lZ / (1+d2lZ/tau_ni)
            ttau = max(ttau,0.1) # enforce positivity i.e. lower bound ttau by zero
            tnu = ( dlZ - nu_ni/tau_ni*d2lZ )/(1+d2lZ/tau_ni);
            q_fir[i] = Dict("mu" => tnu/ttau, "v" => ones(1,1)/ttau)
            #println(i,": yr=", y[i], " q_fir[i]=", q_fir[i])
            lZi_hat_r[i] = lZ[1,1]
            m_fir_gr[i] = q_fir[i]
            m_fir_gr[i], m_fir_gr_old[i] = dampening(m_fir_gr[i], m_fir_gr_old[i], dampening_factor, iter)
            #mm = m_fir_gr[i]; if isnan(mm["mu"][1,1]) || isinf(mm["mu"][1,1]) || isnan(mm["v"][1,1]) || isinf(mm["v"][1,1]); println(i, " m_fir_gr[i]=", mm); print(fail); end
            tmu_r[i] = m_fir_gr[i]["mu"][1,1]
            tsigma_r[i,i] = m_fir_gr[i]["v"][1,1]
            if DEBUG @printf("Step 10,11,12: updated tilde R: m_fir_gr = N(%.3f, %.3f)\n", m_fir_gr[i]["mu"][1,1], m_fir_gr[i]["v"][1,1]); end
            for e=1:size(x_events[i],1)
                e_ix = index_mapping[i][e]
                lZ, dlZ, d2lZ = lik_ep_quad(max(m_yie_hie[i][e]["mu"],0), m_ge_fie[i][e]["mu"], m_ge_fie[i][e]["v"])
                tau_ni = 1.0 / m_ge_fie[i][e]["v"][1,1]
                nu_ni = m_ge_fie[i][e]["mu"][1,1] / m_ge_fie[i][e]["v"][1,1]
                ttau = -d2lZ / (1+d2lZ/tau_ni)
                ttau = max(ttau,0.1) # enforce positivity i.e. lower bound ttau by zero
                tnu = ( dlZ - nu_ni/tau_ni*d2lZ )/(1+d2lZ/tau_ni);
                q_fie[i][e] = Dict("mu" => tnu/ttau, "v" => ones(1,1)/ttau)
                #println(i,": ye=", y[i], " q_fie[i]=", q_fie[i])
                lZi_hat_e[e_ix] = lZ[1,1]
                m_fie_ge[i] = q_fie[i]
                m_fie_ge[i][e], m_fie_ge_old[i][e] = dampening(m_fie_ge[i][e], m_fie_ge_old[i][e], dampening_factor, iter)
                #mm = m_fie_ge[i][e]; if isnan(mm["mu"][1,1]) || isinf(mm["mu"][1,1]) || isnan(mm["v"][1,1]) || isinf(mm["v"][1,1]); println(i, " m_fie_ge[i]=", mm); print(fail); end
                tmu_e[e_ix] = m_fie_ge[i][e]["mu"][1,1]
                tsigma_e[e_ix,e_ix] = m_fie_ge[i][e]["v"][1,1]
                if DEBUG @printf("Step 10,11,12: updated tilde E%d: m_fie_ge = N(%.3f, %.3f)\n", e, m_fie_ge[i][e]["mu"][1,1], m_fie_ge[i][e]["v"][1,1]); end
            end
            
            
            # step 11: compute posterion on f_i variables
            #q_fir[i] = gaussianProduct(m_gr_fir[i], m_hir_fir[i])
            #for e=1:size(x_events[i],1)
            #    q_fie[i][e] = gaussianProduct(m_ge_fie[i][e], m_hie_fie[i][e])
            #end
            #if DEBUG @printf("Step 11: q_fir[i] = N(%.3f, %.3f) \t q_fie[i] = N(%.3f, %.3f)\n", q_fir[i]["mu"], q_fir[i]["v"], q_fie[i]["mu"], q_fie[i]["v"]); end
            

            # step 12: message from f_i variables to g factors
            #m_fir_gr[i] = gaussianDivision(q_fir[i], m_gr_fir[i])
            #m_fir_gr[i], m_fir_gr_old[i] = dampening(m_fir_gr[i], m_fir_gr_old[i], dampening_factor, iter)
            #tmu_r[i] = m_fir_gr[i]["mu"][1,1]
            #tsigma_r[i,i] = m_fir_gr[i]["v"][1,1]
            #tsigma_r_inv[i,i] = 1.0 / (m_fir_gr[i]["v"][1,1] + 0.000000000001)
            #for e=1:size(x_events[i],1)
            #    e_ix = index_mapping[i][e]
            #    m_fie_ge[i][e] = gaussianDivision(q_fie[i][e], m_ge_fie[i][e])
            #    m_fie_ge[i][e], m_fie_ge_old[i][e] = dampening(m_fie_ge[i][e], m_fie_ge_old[i][e], dampening_factor, iter)
                #tmu_e[e_ix] = m_fie_ge[i][e]["mu"][1,1]
                #tsigma_e[e_ix,e_ix] = m_fie_ge[i][e]["v"][1,1]
            #    #tsigma_e_inv[e_ix,e_ix] = 1.0 / (m_fie_ge[i][e]["v"][1,1] + 0.000000000001)
            #end
            #if DEBUG @printf("Step 12: m_fir_gr[i] = N(%.3f, %.3f) \t m_fie_ge[i] = N(%.3f, %.3f)\n", m_fir_gr[i]["mu"], m_fir_gr[i]["v"], m_fie_ge[i]["mu"], m_fie_ge[i]["v"]); end

        end
        
        #tsigma_r_inv = sinv( tsigma_r )
        #q_fr["v"] = sinv( K_routine_inv + tsigma_r_inv )
        q_fr["v"] = K_routine - K_routine * ((tsigma_r + K_routine) \ K_routine)
        q_fr["mu"] = q_fr["v"] * (tsigma_r \ tmu_r)
        
        #tsigma_e_inv = sinv( tsigma_e )
        #q_fe["v"] = sinv( K_events_inv + tsigma_e_inv )
        q_fe["v"] = K_events - K_events * ((tsigma_e + K_events) \ K_events)
        q_fe["mu"] = q_fe["v"] * (tsigma_e \ tmu_e)

        # compute error statistics
        cc, mae, rmse, rae, rrse, r2 = compute_error_stats(y_total, predictive_dist, print=print_debug)
        if !print_debug
            @printf("\r[infer] iter %d: cc=%.3f r2=%.3f          \t", iter, cc, r2)
        end
    end

    posterior = Dict("y"=>predictive_dist, "y_r"=>yir_marginals, "y_e"=>yie_marginals, "K_r"=>K_routine, "q_fr"=>q_fr, "tmu_r"=>tmu_r, "tsigma_r"=>tsigma_r , "tau_ni_r"=>tau_ni_r, "nu_ni_r"=>nu_ni_r, "K_e"=>K_events, "q_fe"=>q_fe, "tmu_e"=>tmu_e, "tsigma_e"=>tsigma_e, "tau_ni_e"=>tau_ni_e, "nu_ni_e"=>nu_ni_e, "q_yir"=>q_yir, "q_yie"=>q_yie, "lZi_hat_r"=>lZi_hat_r, "lZi_hat_e"=>lZi_hat_e, "ix_map"=>index_mapping)

    return posterior
end

function bam_poisson_gp_pred(x_routine_train, x_routine_test, x_events_train, x_events_test, x_events_flat_train, x_events_flat_test, v, posterior, cov_routine, cov_events; print_debug=true)
    DEBUG = false
    N = size(x_routine_test,1)
    D_routine = size(x_routine_test,2)
    D_events = size(x_events_flat_test,2)

    if print_debug
        println("\nrunning expectation propagation for prediction...")
    end

    K_tsigma_r_inv = inv(posterior["K_r"] + posterior["tsigma_r"])
    K_tsigma_e_inv = inv(posterior["K_e"] + posterior["tsigma_e"])

    # data structure for storing the approximate posterior on f
    BigNumber = 10.0^200

    # messages from f_i variables to h_i factors
    m_fir_hir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_fie_hie = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from h_i factors to y_i variables
    m_hir_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hie_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initialize messages and marginals approximations
    for i=1:N
        m_fir_hir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_hir_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_fie_hie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        for e=1:size(x_events_test[i],1)
            push!(m_fie_hie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_hie_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
        end
        push!(m_fie_hie, m_fie_hie_inner)
        push!(m_hie_yie, m_hie_yie_inner)
    end

    # run EP
    yir_marginals = zeros(N,1)
    yie_marginals = Float64[]
    predictive_dist = Dict("mu" => zeros(N,1), "v" => zeros(N,1))
    N_quad = 20
    t,w = gauher(N_quad)
    lw = log(w')
    for i=1:N # sweep through the data points
        K_xstar = covSEard(x_routine_test[i,:], x_routine_train, cov_routine["v"], cov_routine["l"])
        K_xstar_xstar = covSEard(x_routine_test[i,:], x_routine_test[i,:], cov_routine["v"], cov_routine["l"])
        m_fir_hir[i]["mu"] = (K_xstar * K_tsigma_r_inv) * posterior["q_fr"]["mu"]
        m_fir_hir[i]["v"] = K_xstar_xstar - (K_xstar * K_tsigma_r_inv) * K_xstar'
        for e=1:size(x_events_test[i],1)
            K_xstar = covSEard(x_events_test[i][e], x_events_flat_train, cov_events["v"], cov_events["l"])
            K_xstar_xstar = covSEard(x_events_test[i][e], x_events_test[i][e], cov_events["v"], cov_events["l"])
            m_fie_hie[i][e]["mu"] = (K_xstar * K_tsigma_e_inv) * posterior["q_fe"]["mu"]
            m_fie_hie[i][e]["v"] = K_xstar_xstar - (K_xstar * K_tsigma_e_inv) * K_xstar'
        end
        

        # step 4: message from h_i factors to y_i variables
        #q_yir[i] = gaussianMarginal(m_fir_hir[i], 1, 0, beta_routine)
        #for e=1:size(x_events_test[i],1)
        #    q_yie[i][e] = gaussianMarginal(m_fie_hie[i][e], 1, 0, beta_events)
        #end
        
        
        # step 5: compute posterior on y_i^r and y_i^e variables
        #q_yir[i], Z_hat = gaussianTruncation(q_yir[i], 0)
        #yir_marginals[i] = q_yir[i]["mu"][1,1]
        #for e=1:size(x_events_test[i],1)
        #    q_yie[i][e], Z_hat = gaussianTruncation(q_yie[i][e], 0)
        #    push!(yie_marginals, q_yie[i][e]["mu"][1,1])
        #end

        
        # step 4: message from h_i factors to y_i variables
        # first moment using Gaussian-Hermite quad
        mu = m_fir_hir[i]["mu"].*ones(1,1)
        sig = sqrt(m_fir_hir[i]["v"]).*ones(1,1)
        lg = sig*t'+mu*ones(1,N_quad)
        lZ,_ = logsumexp(lg+lw)
        ymu = exp(lZ)
        # second y moment from Poisson distribution
        elg = exp(lg)
        yv = elg
        ys2 = (yv+(elg-ymu*ones(1,N_quad)).^2)*w
        m_hir_yir[i] = Dict("mu" => ones(1,1) * ymu[1,1], "v" => ones(1,1) * ys2[1,1])
        yir_marginals[i] = m_hir_yir[i]["mu"][1,1]
        if DEBUG @printf("Step 4: predictive R: m_hir_yir = N(%.3f, %.3f)\n", m_hir_yir[i]["mu"][1,1], m_hir_yir[i]["v"][1,1]); end
        for e=1:size(x_events_test[i],1)
            # first moment using Gaussian-Hermite quad
            mu = m_fie_hie[i][e]["mu"].*ones(1,1)
            sig = sqrt(m_fie_hie[i][e]["v"]).*ones(1,1)
            lg = sig*t'+mu*ones(1,N_quad)
            lZ,_ = logsumexp(lg+lw)
            ymu = exp(lZ)
            # second y moment from Poisson distribution
            elg = exp(lg)
            yv = elg
            ys2 = (yv+(elg-ymu*ones(1,N_quad)).^2)*w
            m_hie_yie[i][e] = Dict("mu" => ones(1,1) * ymu[1,1], "v" => ones(1,1) * ys2[1,1])
            push!(yie_marginals, m_hie_yie[i][e]["mu"][1,1])
            if DEBUG @printf("Step 4: predictive E%d: m_hie_yie = N(%.3f, %.3f)\n", e, m_hie_yie[i][e]["mu"][1,1], m_hie_yie[i][e]["v"][1,1]); end
        end
        
        
        # compute predictive distribution and the sum of the components distributions
        predictive_dist["mu"][i] = m_hir_yir[i]["mu"][1,1]
        predictive_dist["v"][i] = v + m_hir_yir[i]["v"][1,1]
        for e=1:size(x_events_test[i],1)
            predictive_dist["mu"][i] += m_hie_yie[i][e]["mu"][1,1]
            predictive_dist["v"][i] += m_hie_yie[i][e]["v"][1,1]
        end
    end

    predictive_posterior = Dict("y"=>predictive_dist, "y_r"=>yir_marginals, "y_e"=>yie_marginals)

    return predictive_posterior
end


function bam_poisson_gp_marginal_lik(x_routine, x_events, x_events_flat, y_total, beta_routine, beta_events, v, cov_routine, cov_events; derivatives=false)
    N = size(x_routine,1)
    N_events = size(x_events_flat,1)
    D_routine = size(x_routine,2)
    D_events = size(x_events_flat,2)

    post = Dict()
    println()
    try
        post = bam_poisson_gp_inf(x_routine, x_events, x_events_flat, y_total, v, cov_routine, cov_events, 0; print_debug=false)
    catch
        println("\nwarning: inference method failed")
        num_params = length(cov_routine["l"]) + length(cov_routine["v"]) + length(cov_events["l"]) + length(cov_events["v"]) # cov params
        if derivatives; return [-Inf], Inf*ones(num_params); else; return [-Inf]; end
    end
    println()

    #try
        # routine part of marginal lik
        ttau_r = 1.0 ./ diag(post["tsigma_r"])
        ttau_r = max(ttau_r,0)
        post["tau_ni_r"] = max(post["tau_ni_r"],0)
        tnu_r = ttau_r .* post["tmu_r"]
        sW_r = sqrt(ttau_r)
        L_r = chol(eye(N)+sW_r*sW_r'.*post["K_r"])
        v_r = diag(post["q_fr"]["v"])
        mu_ni_r = post["nu_ni_r"]./post["tau_ni_r"]
        v_ni_r = 1.0./post["tau_ni_r"]
        lZ_r = -sum(log(diag(L_r))) + tnu_r'*post["q_fr"]["v"]*tnu_r/2 - (v_r'*tnu_r.^2)/2
        lZ_r += post["nu_ni_r"]'*((ttau_r./post["tau_ni_r"].*post["nu_ni_r"] - 2*tnu_r).*v_r)/2 
        lZ_r += sum(log(1+ttau_r./post["tau_ni_r"]))/2
        #println("lZ_r before: ", lZ_r)
        for i=1:N
            lZi = post["lZi_hat_r"][i]
            lZ_r += lZi
        end

        # events part of marginal lik
        ttau_e = 1.0 ./ diag(post["tsigma_e"])
        ttau_e = max(ttau_e,0)
        post["tau_ni_e"] = max(post["tau_ni_e"],0)
        tnu_e = ttau_e .* post["tmu_e"]
        sW_e = sqrt(ttau_e)
        L_e = chol(eye(N_events)+sW_e*sW_e'.*post["K_e"])
        v_e = diag(post["q_fe"]["v"])
        mu_ni_e = post["nu_ni_e"]./post["tau_ni_e"]
        v_ni_e = 1.0./post["tau_ni_e"]
        lZ_e = -sum(log(diag(L_e))) + tnu_e'*post["q_fe"]["v"]*tnu_e/2 - (v_e'*tnu_e.^2)/2
        lZ_e += post["nu_ni_e"]'*((ttau_e./post["tau_ni_e"].*post["nu_ni_e"] - 2*tnu_e).*v_e)/2 
        lZ_e += sum(log(1+ttau_e./post["tau_ni_e"]))/2
        #println("lZ_e before: ", lZ_e)
        for i=1:N
            for e=1:size(x_events[i],1)
                e_ix = post["ix_map"][i][e]
                lZi = post["lZi_hat_e"][e_ix]
                lZ_e += lZi
            end
        end

        lZ = lZ_r + lZ_e
        println("lZ_r=", lZ_r, "   lZ_e=", lZ_e, "   lZ=", lZ)

        if derivatives
            num_params = length(cov_routine["l"]) + length(cov_routine["v"]) + length(cov_events["l"]) + length(cov_events["v"]) # cov params
            dlZ = zeros(num_params)

            # covariance hyperparameters
            # routine part
            alpha_r = tnu_r - sW_r.*solve_chol(L_r,sW_r.*(post["K_r"]*tnu_r))
            F_r = alpha_r*alpha_r' - repmat(sW_r,1,N) .* solve_chol(L_r,diagm(sW_r))
            #println("routine cov:")
            for p=1:D_routine
                dK = covSEard(x_routine, x_routine, cov_routine["v"], cov_routine["l"], p=p)
                dlZ[p] = sum(F_r.*dK)/2
            #    print(dlZ[p],", ")
            end
            #println()
            # events part
            alpha_e = tnu_e - sW_e.*solve_chol(L_e,sW_e.*(post["K_e"]*tnu_e))
            F_e = alpha_e*alpha_e' - repmat(sW_e,1,N_events) .* solve_chol(L_e,diagm(sW_e))
            #println("events cov:")
            for p=1:D_events
                dK = covSEard(x_events_flat, x_events_flat, cov_events["v"], cov_events["l"], p=p)
                dlZ[D_routine + 1 + p] = sum(F_e.*dK)/2
            #    print(dlZ[D_routine + 1 + p],", ")
            end
            #println()

            # there are not likelihood parameters for the poisson

            return lZ[1,1], dlZ
        else
            return lZ[1,1]
        end
    #catch
    #    println("\nwarning: marginal lik method failed\n")
    #    if derivatives; return [-Inf], Inf; else; return [-Inf]; end
    #end
end


