
include("covariance_functions.jl")
include("gaussian_ops.jl")
include("utils.jl")

function bam_truncated_linear_gaussian_inf(x_routine, x_events, x_events_flat, y_total, routine_weights_prior, events_weights_prior, beta_routine, beta_events, v; print_debug=true)
    N = size(x_routine,1)
    D_routine = size(routine_weights_prior["mu"],1)
    D_events = size(events_weights_prior["mu"],1)

    if print_debug
        println("\nrunning BAM-LR (Truncated Gaussian)...")
        println("iter 0: initialization")
    end

    # data structures for storing the approximate marginals
    BigNumber = 10.0^200
    q_wr = Dict("nu" => inv(routine_weights_prior["v"]), "tau" => routine_weights_prior["v"] \ routine_weights_prior["mu"], "mu" => routine_weights_prior["mu"], "v" => routine_weights_prior["v"])
    old_q_wr = copy(q_wr)
    q_we = Dict("nu" => inv(events_weights_prior["v"]), "tau" => events_weights_prior["v"] \ events_weights_prior["mu"], "mu" => events_weights_prior["mu"], "v" => events_weights_prior["v"])
    old_q_we = copy(q_we)

    q_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    q_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # messages from priors (we are converting from moment parameters to natural parameters)
    m0_fr_wr = Dict("nu" => inv(routine_weights_prior["v"]), "tau" => routine_weights_prior["v"] \ routine_weights_prior["mu"])
    m0_fe_we = Dict("nu" => inv(events_weights_prior["v"]), "tau" => events_weights_prior["v"] \ events_weights_prior["mu"])

    # initilize messages from w variables to h_i factors
    m_wr_hir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_wr_hir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_we_hie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_we_hie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initilize messages from h_i factors to w variables
    m_hir_wr = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hir_wr_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hie_we = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_hie_we_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initilize messages from h_i factors to y_i^r variables
    m_hir_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hir_yir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hie_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_hie_yie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initilize messages from y_i^r variables to h_i factors
    m_yir_hir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yir_hir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yie_hie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_yie_hie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initilize messages from y_i^r variables to k_i factors
    m_yir_ki = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yir_ki_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yie_ki = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_yie_ki_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initilize messages from k_i factors to y_i^r variables
    m_ki_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_ki_yir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_ki_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_ki_yie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initialize messages and marginals approximations
    for i=1:N
        q_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1)*BigNumber)
        m_wr_hir[i] = Dict("tau" => zeros(D_routine,1), "nu" => zeros(D_routine,D_routine))
        m_wr_hir_old[i] = Dict("tau" => zeros(D_routine,1), "nu" => zeros(D_routine,D_routine))
        m_hir_wr[i] = Dict("tau" => zeros(D_routine,1), "nu" => zeros(D_routine,D_routine))
        m_hir_wr_old[i] = Dict("tau" => zeros(D_routine,1), "nu" => zeros(D_routine,D_routine))
        m_hir_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_hir_yir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_hir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_hir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_ki[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_ki_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_ki_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_ki_yir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)

        q_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_we_hie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_we_hie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_we_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_we_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_yie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_hie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_hie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_ki_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_ki_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_ki_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_ki_yie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        for e=1:size(x_events[i],1)
            push!(q_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1)*BigNumber))
            push!(m_we_hie_inner, Dict("tau" => zeros(D_events,1), "nu" => zeros(D_events,D_events)))
            push!(m_we_hie_old_inner, Dict("tau" => zeros(D_events,1), "nu" => zeros(D_events,D_events)))
            push!(m_hie_we_inner, Dict("tau" => zeros(D_events,1), "nu" => zeros(D_events,D_events)))
            push!(m_hie_we_old_inner, Dict("tau" => zeros(D_events,1), "nu" => zeros(D_events,D_events)))
            push!(m_hie_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_hie_yie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_hie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_hie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_ki_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_ki_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_ki_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_ki_yie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
        end
        push!(q_yie, q_yie_inner)
        push!(m_we_hie, m_we_hie_inner)
        push!(m_we_hie_old, m_we_hie_old_inner)
        push!(m_hie_we, m_hie_we_inner)
        push!(m_hie_we_old, m_hie_we_old_inner)
        push!(m_hie_yie, m_hie_yie_inner)
        push!(m_hie_yie_old, m_hie_yie_old_inner)
        push!(m_yie_hie, m_yie_hie_inner)
        push!(m_yie_hie_old, m_yie_hie_old_inner)
        push!(m_yie_ki, m_yie_ki_inner)
        push!(m_yie_ki_old, m_yie_ki_old_inner)
        push!(m_ki_yie, m_ki_yie_inner)
        push!(m_ki_yie_old, m_ki_yie_old_inner)
    end

    inv_xr = Array{Float64}[]
    inv_xr_br_xr = Array{Float64}[]
    inv_xe = Array{Array{Float64}}[]
    inv_xe_be_xe = Array{Array{Float64}}[]
    for i=1:N
        inv_x = pinv(x_routine[i,:])
        push!(inv_xr, inv_x)
        push!(inv_xr_br_xr, inv_x * beta_routine * inv_x')
        inv_xe_vector = Array{Float64}[]
        inv_xe_be_xe_vector = Array{Float64}[]
        for e=1:size(x_events[i],1)
            inv_x = pinv(x_events[i][e])
            push!(inv_xe_vector, inv_x)
            push!(inv_xe_be_xe_vector, inv_x * beta_events * inv_x')
        end
        push!(inv_xe, inv_xe_vector)
        push!(inv_xe_be_xe, inv_xe_be_xe_vector)
    end

    # run EP
    max_iters = 35
    dampening_factor = 0.7

    routine_weights_posterior = Dict{ASCIIString,Array{Float64,2}}
    events_weights_posterior = Dict{ASCIIString,Array{Float64,2}}
    q_yir_mu = zeros(N,1)
    q_yie_mu = Float64[]
    predictive_dist = Dict("mu" => zeros(N,1))

    for iter=1:max_iters
        if print_debug
            print("iter $iter: ")
        end
        
        q_yir_mu = zeros(N,1)
        q_yie_mu = Float64[]
        predictive_dist = Dict("mu" => zeros(N,1))
        
        for i=1:N # sweep through the data points
            #println("processing data point $i, y($i)=$(y_total[i])")
            
            # step 2: compute the messages from the weights variables down to the h_i factors
            m_wr_hir[i] = gaussianDivision(q_wr, m_hir_wr[i])
            m_wr_hir[i], m_wr_hir_old[i] = dampening(m_wr_hir[i], m_wr_hir_old[i], dampening_factor, iter)
            
            for e=1:size(x_events[i],1)
                m_we_hie[i][e] = gaussianDivision(q_we, m_hie_we[i][e])
                m_we_hie[i][e], m_we_hie_old[i][e] = dampening(m_we_hie[i][e], m_we_hie_old[i][e], dampening_factor, iter)
            end
        

            # step 3: compute the messages from the h_i factors down to the y_i variables
            m_hir_yir[i] = gaussianMarginal(m_wr_hir[i], x_routine[i,:], 0, beta_routine)
            m_hir_yir[i], m_hir_yir_old[i] = dampening(m_hir_yir[i], m_hir_yir_old[i], dampening_factor, iter)
            
            for e=1:size(x_events[i],1)
                m_hie_yie[i][e] = gaussianMarginal(m_we_hie[i][e], x_events[i][e], 0, beta_events)
                m_hie_yie[i][e], m_hie_yie_old[i][e] = dampening(m_hie_yie[i][e], m_hie_yie_old[i][e], dampening_factor, iter)
            end
            

            # step 4: compute the posteriors over the component variables q^t(y_i^r) and q^t(y_i^e)
            p_hat = gaussianProduct(m_hir_yir[i],m_ki_yir[i])
            q_yir[i], Z_hat = gaussianTruncation(p_hat, 0)
            #q_yir[i] = p_hat

            for e=1:size(x_events[i],1)
                p_hat = gaussianProduct(m_hie_yie[i][e],m_ki_yie[i][e])
                q_yie[i][e], Z_hat = gaussianTruncation(p_hat, 0)
                #q_yie[i][e] = p_hat
            end
            
            
            # step 5: compute the messages from the components y_i^r and y_i^e variables down to the k_i factor
            m_yir_ki[i] = gaussianDivision(q_yir[i],m_ki_yir[i])
            m_yir_ki[i], m_yir_ki_old[i] = dampening(m_yir_ki[i], m_yir_ki_old[i], dampening_factor, iter)
            
            for e=1:size(x_events[i],1)
                m_yie_ki[i][e] = gaussianDivision(q_yie[i][e],m_ki_yie[i][e])
                m_yie_ki[i][e], m_yie_ki_old[i][e] = dampening(m_yie_ki[i][e], m_yie_ki_old[i][e], dampening_factor, iter)
            end

            # step 6: compute the upwards messages from the factors k_i
            m_ki_yir[i]["mu"] = y_total[i]*ones(1,1)
            m_ki_yir[i]["v"] = v*ones(1,1)
            for e=1:size(x_events[i],1)
                m_ki_yir[i]["mu"] -= m_yie_ki[i][e]["mu"]
                m_ki_yir[i]["v"] += m_yie_ki[i][e]["v"]
            end
            m_ki_yir[i], m_ki_yir_old[i] = dampening(m_ki_yir[i], m_ki_yir_old[i], dampening_factor, iter)

            for e=1:size(x_events[i],1)
                m_ki_yie[i][e]["mu"] = y_total[i] - m_yir_ki[i]["mu"]
                m_ki_yie[i][e]["v"] = v + m_yir_ki[i]["v"]
                for e2=1:size(x_events[i],1)
                    if e2 != e
                        m_ki_yie[i][e]["mu"] -= m_yie_ki[i][e2]["mu"]
                        m_ki_yie[i][e]["v"] += m_yie_ki[i][e2]["v"]
                    end
                end
                m_ki_yie[i][e], m_ki_yie_old[i][e] = dampening(m_ki_yie[i][e], m_ki_yie_old[i][e], dampening_factor, iter)
            end
            

            # step 7: compute the (revised) posteriors over the component variables q^t(y_i^r) and q^t(y_i^e)
            p_hat = gaussianProduct(m_hir_yir[i],m_ki_yir[i])
            q_yir[i], Z_hat = gaussianTruncation(p_hat, 0)
            #q_yir[i] = p_hat
            q_yir_mu[i] = q_yir[i]["mu"][1,1]
            predictive_dist["mu"][i] = q_yir[i]["mu"][1,1]
            for e=1:size(x_events[i],1)
                p_hat = gaussianProduct(m_hie_yie[i][e],m_ki_yie[i][e])
                q_yie[i][e], Z_hat = gaussianTruncation(p_hat, 0)
                #q_yie[i][e] = p_hat
                push!(q_yie_mu, q_yie[i][e]["mu"][1,1])
                predictive_dist["mu"][i] += q_yie[i][e]["mu"][1,1]
            end
            
            
            # step 8: compute the upwards messages from the components variables y_i^r and y_i^e to the factors h_i^r and h_i^e, respectively
            m_yir_hir[i] = gaussianDivision(q_yir[i], m_hir_yir[i])
            m_yir_hir[i], m_yir_hir_old[i] = dampening(m_yir_hir[i], m_yir_hir_old[i], dampening_factor, iter)
            
            for e=1:size(x_events[i],1)
                m_yie_hie[i][e] = gaussianDivision(q_yie[i][e], m_hie_yie[i][e])
                m_yie_hie[i][e], m_yie_hie_old[i][e] = dampening(m_yie_hie[i][e], m_yie_hie_old[i][e], dampening_factor, iter)
            end

            
            # step 9: compute the upwards messages from the factors h_i to the weights variables w
            m_hir_wr[i] = gaussianMarginal(m_yir_hir[i], inv_xr[i], 0, inv_xr_br_xr[i])
            momentToNaturalParameters!(m_hir_wr[i])
            m_hir_wr[i], m_hir_wr_old[i] = dampening(m_hir_wr[i], m_hir_wr_old[i], dampening_factor, iter)
            
            for e=1:size(x_events[i],1)
                m_hie_we[i][e] = gaussianMarginal(m_yie_hie[i][e], inv_xe[i][e], 0, inv_xe_be_xe[i][e])
                momentToNaturalParameters!(m_hie_we[i][e])
                m_hie_we[i][e], m_hie_we_old[i][e] = dampening(m_hie_we[i][e], m_hie_we_old[i][e], dampening_factor, iter)
            end
            
        end
        
        # step 1: compute posterior over the weights variables
        q_wr = deepcopy(m0_fr_wr)
        q_we = deepcopy(m0_fe_we) 
        for j=1:N
            q_wr = gaussianProduct(q_wr, m_hir_wr[j])
            for e=1:size(x_events[j],1)
                q_we = gaussianProduct(q_we, m_hie_we[j][e])
            end
        end
        naturalToMomentParameters!(q_wr)
        naturalToMomentParameters!(q_we)
        
        routine_weights_posterior = Dict("mu" => q_wr["mu"], "v" => q_wr["v"])
        events_weights_posterior = Dict("mu" => q_we["mu"], "v" => q_we["v"])
        
        # compute error statistics
        cc, mae, rmse, rae, rrse = compute_error_stats(y_total, predictive_dist, print=print_debug)
    end

    posterior = Dict("y"=>predictive_dist, "w_r"=>routine_weights_posterior, "w_e"=>events_weights_posterior, "q_yir_mu"=>q_yir_mu, "q_yie_mu"=>q_yie_mu, "q_yir"=>q_yir, "q_yie"=>q_yie)

    return posterior
end

function bam_truncated_linear_gaussian_pred(x_routine, x_events, x_events_flat, y_total, posterior, beta_routine, beta_events, v; print_debug=true)
    N = size(x_routine,1)
    D_routine = size(posterior["w_r"]["mu"],1)
    D_events = size(posterior["w_e"]["mu"],1)

    # data structures for storing the approximate marginals
    BigNumber = 10.0^200
    #q_wr = ["nu" => inv(posterior["w_r"]["v"]), "tau" => posterior["w_r"]["v"] \ posterior["w_r"]["mu"], "mu" => posterior["w_r"]["mu"], "v" => posterior["w_r"]["v"]]
    q_wr = Dict("nu" => pinv(posterior["w_r"]["v"]), "tau" => pinv(posterior["w_r"]["v"]) * posterior["w_r"]["mu"], "mu" => posterior["w_r"]["mu"], "v" => posterior["w_r"]["v"])
    #q_we = ["nu" => inv(posterior["w_e"]["v"]), "tau" => posterior["w_e"]["v"] \ posterior["w_e"]["mu"], "mu" => posterior["w_e"]["mu"], "v" => posterior["w_e"]["v"]]
    q_we = Dict("nu" => pinv(posterior["w_e"]["v"]), "tau" => pinv(posterior["w_e"]["v"]) * posterior["w_e"]["mu"], "mu" => posterior["w_e"]["mu"], "v" => posterior["w_e"]["v"])
    
    q_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    q_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    q_yi = Array(Dict{ASCIIString,Array{Float64,2}},N)

    # initilize messages from h_i factors to y_i^r variables
    m_hir_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hir_yir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_hie_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_hie_yie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initilize messages from y_i^r variables to h_i factors
    m_yir_hir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yir_hir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yie_hie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_yie_hie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initilize messages from y_i^r variables to k_i factors
    m_yir_ki = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yir_ki_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yie_ki = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_yie_ki_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initilize messages from k_i factors to y_i^r variables
    m_ki_yir = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_ki_yir_old = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_ki_yie = Array{Dict{ASCIIString,Array{Float64,2}}}[]
    m_ki_yie_old = Array{Dict{ASCIIString,Array{Float64,2}}}[]

    # initilize messages from y_i variables to k_i factors
    m_yi_ki = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_yi_ki_old = Array(Dict{ASCIIString,Array{Float64,2}},N)

    # initilize messages from k_i factors to y_i variables
    m_ki_yi = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_ki_yi_old = Array(Dict{ASCIIString,Array{Float64,2}},N)

    # initialize messages and marginals approximations
    for i=1:N
        q_yi[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1)*BigNumber)
        m_yi_ki[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yi_ki_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_ki_yi[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_ki_yi_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)

        q_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1)*BigNumber)
        m_hir_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_hir_yir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_hir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_hir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_ki[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_yir_ki_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_ki_yir[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)
        m_ki_yir_old[i] = Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber)

        q_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_hie_yie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_hie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_hie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_ki_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_yie_ki_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_ki_yie_inner = Dict{ASCIIString,Array{Float64,2}}[]
        m_ki_yie_old_inner = Dict{ASCIIString,Array{Float64,2}}[]
        for e=1:size(x_events[i],1)
            push!(q_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1)*BigNumber))
            push!(m_hie_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_hie_yie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_hie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_hie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_ki_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_yie_ki_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_ki_yie_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
            push!(m_ki_yie_old_inner, Dict("mu" => zeros(1,1), "v" => ones(1,1) * BigNumber))
        end
        push!(q_yie, q_yie_inner)
        push!(m_hie_yie, m_hie_yie_inner)
        push!(m_hie_yie_old, m_hie_yie_old_inner)
        push!(m_yie_hie, m_yie_hie_inner)
        push!(m_yie_hie_old, m_yie_hie_old_inner)
        push!(m_yie_ki, m_yie_ki_inner)
        push!(m_yie_ki_old, m_yie_ki_old_inner)
        push!(m_ki_yie, m_ki_yie_inner)
        push!(m_ki_yie_old, m_ki_yie_old_inner)
    end

    # run EP
    max_iters = 3
    dampening_factor = 1.0

    yir_marginals = zeros(N,1)
    yie_marginals = Float64[]
    yi_marginals = zeros(N,1)
    predictive_dist_test = Dict("mu" => zeros(N,1))

    for iter=1:max_iters
        #println("iteration=$iter")
        #println("iter=", iter, " size=", size(predictive_dists_test["mu"]))

        yie_marginals = Float64[]
        
        for i=1:N # sweep through the data points
            #println("processing data point $i, y($i)=$(y_total[i])")

            # step 2: compute message from y_i to k_i
            m_yi_ki[i] = gaussianDivision(q_yi[i], m_ki_yi[i])
            m_yi_ki[i], m_yi_ki_old[i] = dampening(m_yi_ki[i], m_yi_ki_old[i], dampening_factor, iter)


            # step 3: message from k_i to y_i_r and y_i_e
            m_ki_yir[i]["mu"] = q_yi[i]["mu"]
            m_ki_yir[i]["v"] = v + q_yi[i]["v"]
            for e=1:size(x_events[i],1)
                m_ki_yir[i]["mu"] -= m_yie_ki[i][e]["mu"]
                m_ki_yir[i]["v"] += m_yie_ki[i][e]["v"]
            end
            m_ki_yir[i], m_ki_yir_old[i] = dampening(m_ki_yir[i], m_ki_yir_old[i], dampening_factor, iter)

            for e=1:size(x_events[i],1)
                m_ki_yie[i][e]["mu"] = q_yi[i]["mu"] - m_yir_ki[i]["mu"]
                m_ki_yie[i][e]["v"] = v + q_yi[i]["v"] + m_yir_ki[i]["v"]
                for j=1:size(x_events[i],1)
                    if j != e
                        m_ki_yie[i][e]["mu"] -= m_yie_ki[i][j]["mu"]
                        m_ki_yie[i][e]["v"] += m_yie_ki[i][j]["v"]
                    end
                end
                m_ki_yie[i][e], m_ki_yie_old[i][e] = dampening(m_ki_yie[i][e], m_ki_yie_old[i][e], dampening_factor, iter)
            end


            # step 4: posterior on y_i_r and y_i_e
            q_yir[i] = gaussianProduct(m_hir_yir[i], m_ki_yir[i])
            q_yir[i], Z_hat = gaussianTruncation(q_yir[i], 0)

            for e=1:size(x_events[i],1)
                q_yie[i][e] = gaussianProduct(m_hie_yie[i][e], m_ki_yie[i][e])
                q_yie[i][e], Z_hat = gaussianTruncation(q_yie[i][e], 0)
            end


            # step 5: message from h_i_r to y_i_r
            m_hir_yir[i] = gaussianMarginal(posterior["w_r"], x_routine[i,:], 0, beta_routine)
            m_hir_yir[i], m_hir_yir_old[i] = dampening(m_hir_yir[i], m_hir_yir_old[i], dampening_factor, iter)
            
            for e=1:size(x_events[i],1)
                m_hie_yie[i][e] = gaussianMarginal(posterior["w_e"], x_events[i][e], 0, beta_events)
                m_hie_yie[i][e], m_hie_yie_old[i][e] = dampening(m_hie_yie[i][e], m_hie_yie_old[i][e], dampening_factor, iter)
            end


            # step 6: posterior on y_i_r and y_i_e
            q_yir[i] = gaussianProduct(m_hir_yir[i], m_ki_yir[i])
            q_yir[i], Z_hat = gaussianTruncation(q_yir[i], 0)
            yir_marginals[i] = q_yir[i]["mu"][1,1]
            predictive_dist_test["mu"][i] = q_yir[i]["mu"][1,1]

            for e=1:size(x_events[i],1)
                q_yie[i][e] = gaussianProduct(m_hie_yie[i][e], m_ki_yie[i][e])
                q_yie[i][e], Z_hat = gaussianTruncation(q_yie[i][e], 0)
                push!(yie_marginals, q_yie[i][e]["mu"][1,1])
                predictive_dist_test["mu"][i] += q_yie[i][e]["mu"][1,1]
            end


            # step 7: message from y_i_r to k_i
            m_yir_ki[i] = gaussianDivision(q_yir[i], m_ki_yir[i])
            m_yir_ki[i], m_yir_ki_old[i] = dampening(m_yir_ki[i], m_yir_ki_old[i], dampening_factor, iter)

            for e=1:size(x_events[i],1)
                m_yie_ki[i][e] = gaussianDivision(q_yie[i][e], m_ki_yie[i][e])
                m_yie_ki[i][e], m_yie_ki_old[i][e] = dampening(m_yie_ki[i][e], m_yie_ki_old[i][e], dampening_factor, iter)
            end


            # step 8: message from k_i to y_i
            m_ki_yi[i]["mu"] = m_yir_ki[i]["mu"]
            m_ki_yi[i]["v"] = v + m_yir_ki[i]["v"]
            for e=1:size(x_events[i],1)
                m_ki_yi[i]["mu"] += m_yie_ki[i][e]["mu"]
                m_ki_yi[i]["v"] += m_yie_ki[i][e]["v"]
            end
            m_ki_yi[i], m_ki_yi_old[i] = dampening(m_ki_yi[i], m_ki_yi_old[i], dampening_factor, iter)

            # posterior on y_i
            q_yi[i] = deepcopy(m_ki_yi[i])
            yi_marginals[i] = q_yi[i]["mu"][1,1]
        end
    end

    predictive_posterior = Dict("y"=>predictive_dist_test, "y_r"=>yir_marginals, "y_e"=>yie_marginals)

    return predictive_posterior
end

function bam_truncated_linear_gaussian_marginal_lik(x_routine, xx_routine, x_events_sum, xx_events_sum, y_total, prior_routine, prior_events, beta_routine, beta_events, v; derivatives=false)
    N = size(x_routine,1)
    D_routine = size(x_routine,2)
    D_events = size(x_events_sum,2)

    C = (v+beta_routine+beta_events)*eye(N) + prior_routine*xx_routine + prior_events*xx_events_sum
    lZ = 0
    L = NaN; invC = NaN; invCy = NaN
    try
        L = chol(C)
        #invC = inv(C)
        invC = L\(L\eye(N))'
        invCy = invC*y_total
        lZ = -N*log(2*pi)/2 -sum(diag(L)) -y_total'*invCy/2
        #lZ = -N*log(2*pi)/2 -log(det(C))/2 -y_total'*invCy/2
    catch
        println("warning: inference method failed")
        if derivatives; return -Inf, [Inf,Inf,Inf,Inf]; else; return -Inf; end
    end

    if derivatives
        dlZ = zeros(4)
        invCxx_r = invC*xx_routine
        invCxx_e = invC*xx_events_sum
        dlZ[1] = -trace(invCxx_r)/2 + sum(y_total'*invCxx_r*invCy)/2
        dlZ[2] = -trace(invCxx_e)/2 + sum(y_total'*invCxx_e*invCy)/2
        dlZ[3] = -trace(invC)/2 + sum(y_total'*invC*invCy)/2
        dlZ[4] = -trace(invC)/2 + sum(y_total'*invC*invCy)/2

        return lZ[1,1], dlZ
    end

    return lZ[1,1]
end

function bam_lr_optimize_hyperparams(x_routine, x_events_sum, y_total, prior_routine, prior_events, beta_routine, beta_events, v; max_iterations=100)
    global gX_routine = x_routine
    global gXX_routine = x_routine*x_routine'
    global gX_events_sum = x_events_sum
    global gXX_events_sum = x_events_sum*x_events_sum'
    global gy_total = y_total
    global gV = v
    global iter = 0
    init_params = [prior_routine, prior_events, beta_routine, beta_events]
    #d = DifferentiableFunction(lr_f)
    d = DifferentiableFunction(bam_lr_f, bam_lr_g!, bam_lr_fg!)
    println("optimizing hyper-parameters...")
    res = Optim.optimize(d, init_params, method=:cg, iterations=max_iterations)
    println()
    nlZ = res.f_minimum
    opt_params = res.minimum
    println("minimum lZ: ", -nlZ)
    println("num iterations: ", res.iterations)
    weights_prior_routine = Dict("mu" => zeros(D_routine,1), "v" => opt_params[1]*eye(D_routine))
    weights_prior_events = Dict("mu" => zeros(D_events,1), "v" => opt_params[2]*eye(D_events))
    beta_routine = opt_params[3]
    beta_events = opt_params[4]
    return weights_prior_routine, weights_prior_events, beta_routine, beta_events, -nlZ[1,1]
end

function bam_lr_f(x::Vector)
    global gX_routine, gXX_routine, gX_events_sum, gXX_events_sum, gy_total, gV, iter
    iter += 1
    prior_routine = x[1]
    prior_events = x[2]
    beta_routine = x[3]
    beta_events = x[4]
    lZ = bam_truncated_linear_gaussian_marginal_lik(gX_routine, gXX_routine, gX_events_sum, gXX_events_sum, gy_total, prior_routine, prior_events, beta_routine, beta_events, gV; derivatives=false)
    #println("\r[eval $iter] lZ: $lZ       \t", x)
    print("\r[eval $iter] lZ: $lZ       \t")
    return -lZ
end

function bam_lr_g!(x::Vector, storage::Vector)
    global gX_routine, gXX_routine, gX_events_sum, gXX_events_sum, gy_total, gV, iter
    prior_routine = x[1]
    prior_events = x[2]
    beta_routine = x[3]
    beta_events = x[4]
    lZ, dlZ = bam_truncated_linear_gaussian_marginal_lik(gX_routine, gXX_routine, gX_events_sum, gXX_events_sum, gy_total, prior_routine, prior_events, beta_routine, beta_events, gV; derivatives=true)
    storage[:] = -dlZ[:]
end

function bam_lr_fg!(x::Vector, storage::Vector)
    global gX_routine, gXX_routine, gX_events_sum, gXX_events_sum, gy_total, gV, iter
    iter += 1
    prior_routine = x[1]
    prior_events = x[2]
    beta_routine = x[3]
    beta_events = x[4]
    lZ, dlZ = bam_truncated_linear_gaussian_marginal_lik(gX_routine, gXX_routine, gX_events_sum, gXX_events_sum, gy_total, prior_routine, prior_events, beta_routine, beta_events, gV; derivatives=true)
    #println("\r[eval $iter] lZ: $lZ       \t", x)
    print("\r[eval $iter] lZ: $lZ       \t")
    storage[:] = -dlZ[:]
    return -lZ
end

