using Optim

function bam_optimize_hyperparams(x_routine, x_events, x_events_flat, y_total, beta_routine, beta_events, v, cov_routine, cov_events, marginal_lik_func; max_iterations=100)
    global gX_routine = x_routine
    global gX_events = x_events
    global gX_events_flat = x_events_flat
    global gy_total = y_total
    global gV = v
    global g_marginal_lik = marginal_lik_func
    global iter = 0

    # setup
    D_routine = size(x_routine,2)
    D_events = size(x_events_flat,2)
    num_params = length(cov_routine["l"]) + length(cov_routine["v"]) + length(cov_events["l"]) + length(cov_events["v"]) + 2 # cov params + lik param
    init_params = zeros(num_params)
    init_params[1:D_routine] = cov_routine["l"]
    init_params[D_routine+1] = cov_routine["v"]
    init_params[D_routine+2:D_routine+1+D_events] = cov_events["l"]
    init_params[D_routine+1+D_events+1] = cov_events["v"]
    init_params[end-1] = beta_routine
    init_params[end] = beta_events

    # do optimization
    #d = DifferentiableFunction(bam_f)
    d = DifferentiableFunction(bam_f, bam_g!, bam_fg!)
    println("optimizing hyper-parameters...")
    res = Optim.optimize(d, init_params, method=:cg, iterations=max_iterations)
    println()
    nlZ = res.f_minimum
    opt_params = res.minimum
    println("minimum lZ: ", -nlZ)
    println("num iterations: ", res.iterations)

    # get results
    cov_routine = Dict("v" => opt_params[D_routine+1], "l" => opt_params[1:D_routine])
    cov_events = Dict("v" => opt_params[D_routine+1+D_events+1], "l" => opt_params[D_routine+2:D_routine+1+D_events])
    beta_routine = opt_params[end-1]
    beta_events = opt_params[end]

    return cov_routine, cov_events, beta_routine, beta_events, -nlZ
end

function bam_f(x::Vector)
    global gX_routine, gX_events, gX_events_flat, gy_total, gV, g_marginal_lik, iter
    iter += 1
    D_routine = size(gX_routine,2)
    D_events = size(gX_events_flat,2)
    cov_routine = Dict("v" => x[D_routine+1], "l" => x[1:D_routine])
    cov_events = Dict("v" => x[D_routine+1+D_events+1], "l" => x[D_routine+2:D_routine+1+D_events])
    beta_routine = x[end-1]
    beta_events = x[end]
    #println("\ncheck:")
    #println("cov routine: ", cov_routine)
    #println("cov events: ", cov_events)
    #if sigma2 > 999999; return Inf; end
    lZ = g_marginal_lik(gX_routine, gX_events, gX_events_flat, gy_total, beta_routine, beta_events, gV, cov_routine, cov_events; derivatives=false)
    print("\r[eval $iter] lZ: $(lZ[1,1])       \t")
    return -lZ[1,1]
end

function bam_g!(x::Vector, storage::Vector)
    global gX_routine, gX_events, gX_events_flat, gy_total, gV, g_marginal_lik
    D_routine = size(gX_routine,2)
    D_events = size(gX_events_flat,2)
    cov_routine = Dict("v" => x[D_routine+1], "l" => x[1:D_routine])
    cov_events = Dict("v" => x[D_routine+1+D_events+1], "l" => x[D_routine+2:D_routine+1+D_events])
    beta_routine = x[end-1]
    beta_events = x[end]
    lZ, dlZ = g_marginal_lik(gX_routine, gX_events, gX_events_flat, gy_total, beta_routine, beta_events, gV, cov_routine, cov_events; derivatives=true)
    storage[:] = -dlZ
end

function bam_fg!(x::Vector, storage::Vector)
    global gX_routine, gX_events, gX_events_flat, gy_total, gV, g_marginal_lik
    D_routine = size(gX_routine,2)
    D_events = size(gX_events_flat,2)
    cov_routine = Dict("v" => x[D_routine+1], "l" => x[1:D_routine])
    cov_events = Dict("v" => x[D_routine+1+D_events+1], "l" => x[D_routine+2:D_routine+1+D_events])
    beta_routine = x[end-1]
    beta_events = x[end]
    lZ, dlZ = g_marginal_lik(gX_routine, gX_events, gX_events_flat, gy_total, beta_routine, beta_events, gV, cov_routine, cov_events; derivatives=true)
    print("\r[eval $iter] lZ: $(lZ[1,1])       \t")
    storage[:] = -dlZ
    return -lZ[1,1]
end


function bam_optimize_cov_hyperparams(x_routine, x_events, x_events_flat, y_total, beta_routine, beta_events, v, cov_routine, cov_events, marginal_lik_func; max_iterations=100)
    global gX_routine = x_routine
    global gX_events = x_events
    global gX_events_flat = x_events_flat
    global gy_total = y_total
    global gBeta_routine = beta_routine
    global gBeta_events = beta_events
    global gV = v
    global g_marginal_lik = marginal_lik_func
    global iter = 0

    # setup
    D_routine = size(x_routine,2)
    D_events = size(x_events_flat,2)
    num_params = length(cov_routine["l"]) + length(cov_routine["v"]) + length(cov_events["l"]) + length(cov_events["v"]) # cov params
    init_params = zeros(num_params)
    init_params[1:D_routine] = cov_routine["l"]
    init_params[D_routine+1] = cov_routine["v"]
    init_params[D_routine+2:end-1] = cov_events["l"]
    init_params[end] = cov_events["v"]

    # do optimization
    #d = DifferentiableFunction(bam_f_cov)
    d = DifferentiableFunction(bam_f_cov, bam_g_cov!, bam_fg_cov!)
    println("optimizing hyper-parameters...")
    res = Optim.optimize(d, init_params, method=:cg, iterations=max_iterations)
    println()
    nlZ = res.f_minimum
    opt_params = res.minimum
    println("minimum lZ: ", -nlZ)
    println("num iterations: ", res.iterations)

    # get results
    cov_routine = Dict("v" => opt_params[D_routine+1], "l" => opt_params[1:D_routine])
    cov_events = Dict("v" => opt_params[end], "l" => opt_params[D_routine+2:end-1])

    return cov_routine, cov_events, beta_routine, beta_events, -nlZ
end

function bam_f_cov(x::Vector)
    global gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, g_marginal_lik, iter
    iter += 1
    D_routine = size(gX_routine,2)
    D_events = size(gX_events_flat,2)
    cov_routine = Dict("v" => x[D_routine+1], "l" => x[1:D_routine])
    cov_events = Dict("v" => x[end], "l" => x[D_routine+2:end-1])
    #println("\ncheck:")
    #println("cov routine: ", cov_routine)
    #println("cov events: ", cov_events)
    #if sigma2 > 999999; return Inf; end
    lZ = g_marginal_lik(gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, cov_routine, cov_events; derivatives=false)
    print("\r[eval $iter] lZ: $(lZ[1,1])       \t")
    return -lZ[1,1]
end

function bam_g_cov!(x::Vector, storage::Vector)
    global gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, g_marginal_lik
    D_routine = size(gX_routine,2)
    D_events = size(gX_events_flat,2)
    cov_routine = Dict("v" => x[D_routine+1], "l" => x[1:D_routine])
    cov_events = Dict("v" => x[end], "l" => x[D_routine+2:end-1])
    lZ, dlZ = g_marginal_lik(gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, cov_routine, cov_events; derivatives=true)
    storage[:] = -dlZ[1:end-2]
end

function bam_fg_cov!(x::Vector, storage::Vector)
    global gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, g_marginal_lik
    D_routine = size(gX_routine,2)
    D_events = size(gX_events_flat,2)
    cov_routine = Dict("v" => x[D_routine+1], "l" => x[1:D_routine])
    cov_events = Dict("v" => x[end], "l" => x[D_routine+2:end-1])
    lZ, dlZ = g_marginal_lik(gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, cov_routine, cov_events; derivatives=true)
    print("\r[eval $iter] lZ: $(lZ[1,1])       \t")
    storage[:] = -dlZ[1:end-2]
    return -lZ[1,1]
end


function bam_optimize_cov_hyperparams(x_routine, x_events, x_events_flat, y_total, beta_routine, beta_events, v, cov_routine, cov_events, marginal_lik_func; max_iterations=100)
    global gX_routine = x_routine
    global gX_events = x_events
    global gX_events_flat = x_events_flat
    global gy_total = y_total
    global gBeta_routine = beta_routine
    global gBeta_events = beta_events
    global gV = v
    global g_marginal_lik = marginal_lik_func
    global iter = 0

    # setup
    D_routine = size(x_routine,2)
    D_events = size(x_events_flat,2)
    num_params = length(cov_routine["l"]) + length(cov_routine["v"]) + length(cov_events["l"]) + length(cov_events["v"]) # cov params
    init_params = zeros(num_params)
    init_params[1:D_routine] = cov_routine["l"]
    init_params[D_routine+1] = cov_routine["v"]
    init_params[D_routine+2:end-1] = cov_events["l"]
    init_params[end] = cov_events["v"]

    # do optimization
    #d = DifferentiableFunction(bam_f_cov)
    d = DifferentiableFunction(bam_f_cov, bam_g_cov!, bam_fg_cov!)
    println("optimizing hyper-parameters...")
    res = Optim.optimize(d, init_params, method=:cg, iterations=max_iterations)
    println()
    nlZ = res.f_minimum
    opt_params = res.minimum
    println("minimum lZ: ", -nlZ)
    println("num iterations: ", res.iterations)

    # get results
    cov_routine = Dict("v" => opt_params[D_routine+1], "l" => opt_params[1:D_routine])
    cov_events = Dict("v" => opt_params[end], "l" => opt_params[D_routine+2:end-1])

    return cov_routine, cov_events, beta_routine, beta_events, -nlZ
end

function bam_f_cov(x::Vector)
    global gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, g_marginal_lik, iter
    iter += 1
    D_routine = size(gX_routine,2)
    D_events = size(gX_events_flat,2)
    cov_routine = Dict("v" => x[D_routine+1], "l" => x[1:D_routine])
    cov_events = Dict("v" => x[end], "l" => x[D_routine+2:end-1])
    #println("\ncheck:")
    #println("cov routine: ", cov_routine)
    #println("cov events: ", cov_events)
    #if sigma2 > 999999; return Inf; end
    lZ = g_marginal_lik(gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, cov_routine, cov_events; derivatives=false)
    print("\r[eval $iter] lZ: $(lZ[1,1])       \t")
    return -lZ[1,1]
end

function bam_g_cov!(x::Vector, storage::Vector)
    global gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, g_marginal_lik
    D_routine = size(gX_routine,2)
    D_events = size(gX_events_flat,2)
    cov_routine = Dict("v" => x[D_routine+1], "l" => x[1:D_routine])
    cov_events = Dict("v" => x[end], "l" => x[D_routine+2:end-1])
    lZ, dlZ = g_marginal_lik(gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, cov_routine, cov_events; derivatives=true)
    storage[:] = -dlZ[1:D_routine+1+D_events+1]
end

function bam_fg_cov!(x::Vector, storage::Vector)
    global gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, g_marginal_lik
    D_routine = size(gX_routine,2)
    D_events = size(gX_events_flat,2)
    cov_routine = Dict("v" => x[D_routine+1], "l" => x[1:D_routine])
    cov_events = Dict("v" => x[end], "l" => x[D_routine+2:end-1])
    lZ, dlZ = g_marginal_lik(gX_routine, gX_events, gX_events_flat, gy_total, gBeta_routine, gBeta_events, gV, cov_routine, cov_events; derivatives=true)
    print("\r[eval $iter] lZ: $(lZ[1,1])       \t")
    storage[:] = -dlZ[1:D_routine+1+D_events+1]
    return -lZ[1,1]
end
