using Optim

function gp_optimize_hyperparams(X, y, cov, sigma, marginal_lik_func; max_iterations=100)
    global gX = X
    global gy = y
    global g_marginal_lik = marginal_lik_func
    global iter = 0

    # setup
    num_params = length(cov["l"]) + length(cov["v"]) + 1 # cov params + lik param
    init_params = zeros(num_params)
    init_params[1:end-2] = cov["l"]
    init_params[end-1] = cov["v"]
    init_params[end] = sigma

    # do optimization
    #d = DifferentiableFunction(gp_f)
    d = DifferentiableFunction(gp_f, gp_g!, gp_fg!)
    println("optimizing hyper-parameters...")
    res = Optim.optimize(d, init_params, method=:cg, iterations=max_iterations)
    println()
    nlZ = res.f_minimum
    opt_params = res.minimum
    println("minimum lZ: ", -nlZ)
    println("num iterations: ", res.iterations)

    # get results
    cov = Dict("v" => opt_params[end-1], "l" => opt_params[1:end-2])
    sigma = opt_params[end]
    
    return cov, sigma, -nlZ
end

function gp_f(x::Vector)
    global gX, gy, g_marginal_lik, iter
    iter += 1
    cov = Dict("v" => x[end-1], "l" => x[1:end-2])
    sigma = x[end]
    #if sigma > 999999; return Inf; end
    lZ = g_marginal_lik(gX, gy, cov, sigma)
    print("\r[eval $iter] lZ: $(lZ[1,1])       \t")
    return -lZ[1,1]
end

function gp_g!(x::Vector, storage::Vector)
    global gX, gy, g_marginal_lik
    cov = Dict("v" => x[end-1], "l" => x[1:end-2])
    sigma = x[end]
    lZ, dlZ = g_marginal_lik(gX, gy, cov, sigma, derivatives=true)
    storage[:] = -dlZ
end

function gp_fg!(x::Vector, storage::Vector)
    global gX, gy, g_marginal_lik, iter
    cov = Dict("v" => x[end-1], "l" => x[1:end-2])
    sigma = x[end]
    lZ, dlZ = g_marginal_lik(gX, gy, cov, sigma, derivatives=true)
    print("\r[eval $iter] lZ: $(lZ[1,1])       \t")
    storage[:] = -dlZ
    return -lZ[1,1]
end
